//  This file is part of the computer game Kartofel.
//  Copyright (C) 2008 Pawel Aleksander Fedorynski <pfedor@fuw.edu.pl>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <cassert>
#include <cerrno>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>

#include "game.h"
#include "trig_table.h"
#include "util.h"

using namespace std;

namespace kartofel {

void LoadLevel(const string& filename, Level* level)
{
  level->lines.Clear();
  level->dots.clear();
  level->starting_phi = 0;
  level->multiplier = 1;
  level->max_time = DEFAULT_MAX_TIME;
  // Add the game field borders.
  level->lines.SetStart(0, 0);
  level->lines.LineTo(0, 10000);
  level->lines.LineTo(10000, 10000);
  level->lines.LineTo(10000, 0);
  level->lines.LineTo(0, 0);

  FILE* f = fopen(filename.c_str(), "r");
  if (!f) {
    cerr << filename << ": " << strerror(errno) << endl;
    exit(-1);
  }
  int line_num = 0;
  char line[1000];
  errno = 0;
  int dot_nr = 1;
  int last_dot_index = -1;
  while (fgets(line, sizeof(line), f)) {
    line_num++;
    int len = strlen(line);
    if (line[0] == '#'  ||  strspn(line, " \t\r\n") == len) {
      continue;
    }
    int x, y, r, phi, max_time, multiplier;
    int n = -1;
    if (sscanf(line, "LineStart %d %d %n", &x, &y, &n), n == len) {
      level->lines.SetStart(x, y);
    } else if (sscanf(line, "LineTo %d %d %n", &x, &y, &n), n == len) {
      level->lines.LineTo(x, y);
    } else if (sscanf(line, "Dot %d %d %d %n", &x, &y, &r, &n), n == len) {
      Dot dot = { x, y, r, REGULAR, dot_nr == 1, {{ dot_nr++, false }} };
      last_dot_index = level->dots.size();
      level->dots.push_back(dot);
    } else if (sscanf(line, "ExtraLife %d %d %d %n", &x, &y, &r, &n),
               n == len) {
      Dot dot = { x, y, r, EXTRA_LIFE, false, {{ false }} };
      level->dots.push_back(dot);
    } else if (sscanf(line, "Phi %d %n", &phi, &n), n == len) {
      level->starting_phi = phi;
    } else if (sscanf(line, "MaxTime %d %n", &max_time, &n), n == len) {
      level->max_time = max_time;
    } else if (sscanf(line, "Multiplier %d %n", &multiplier, &n), n == len) {
      level->multiplier = multiplier;
    } else {
      cerr << filename << ": Error in line " << line_num << endl;
      exit(-1);
    }
  }
  if (ferror(f)) {
    cerr << filename << ": " << strerror(errno) << endl;
    exit(-1);
  }
  if (0 != fclose(f)) {
    cerr << filename << ": " << strerror(errno) << endl;
    exit(-1);
  }
  if (last_dot_index == -1) {
    cerr << filename << ": " << "No dots.\n";
    exit(-1);
  }
  level->dots[last_dot_index].data.regular.last = true;
}

Game::Game()
  : current_dot_(0),
    last_regular_dot_(0),
    waiting_(true),
    current_timestamp_(0),
    score_(0),
    lives_(STARTING_LIVES),
    x_(0), y_(0), phi_(0), speed_(MIN_SPEED) { }

void Game::AddActionHideStarter(vector<Action>* actions)
{
  Action a;
  a.type = HIDE_STARTER;
  actions->push_back(a);
}

void Game::AddActionShowStarter(vector<Action>* actions,
                                int dot_index, int phi)
{
  Action a;
  a.type = SHOW_STARTER;
  a.data.starter.dot_index = dot_index;
  a.data.starter.phi = phi;
  actions->push_back(a);
}

void Game::AddActionDraw(vector<Action>* actions, int start_x, int start_y,
                         int end_x, int end_y)
{
  Action a;
  a.type = DRAW;
  a.data.draw.start_x = start_x;
  a.data.draw.start_y = start_y;
  a.data.draw.end_x = end_x;
  a.data.draw.end_y = end_y;
  actions->push_back(a);
}

void Game::AddActionExplosion(vector<Action>* actions, int x, int y)
{
  Action a;
  a.type = EXPLOSION;
  a.data.explosion.x = x;
  a.data.explosion.y = y;
  actions->push_back(a);
}

void Game::AddActionLightDot(vector<Action>* actions, int dot_index)
{
  Action a;
  a.type = LIGHT_DOT;
  a.data.dot.dot_index = dot_index;
  actions->push_back(a);
}

void Game::AddActionEndLevel(vector<Action>* actions)
{
  Action a;
  a.type = END_LEVEL;
  actions->push_back(a);
}

void Game::AddActionOutOfTime(vector<Action>* actions)
{
  Action a;
  a.type = OUT_OF_TIME;
  actions->push_back(a);
}
  
void Game::AdjustSpeedAndPhi()
{
  if (acc_input_.accelerate_cnt % 4 == 1) {
    speed_ += STEP_SPEED;
    if (speed_ > MAX_SPEED) {
      speed_ = MAX_SPEED;
    }
  }
  if (acc_input_.deccelerate_cnt % 4 == 1) {
    speed_ -= STEP_SPEED;
    if (speed_ < MIN_SPEED) {
      speed_ = MIN_SPEED;
    }
  }
  int step_phi = waiting_ ? STEP_PHI_STARTER : STEP_PHI;
  if (acc_input_.left_cnt) {
    if (!waiting_) {
      step_phi =
        acc_input_.left_cnt < step_phi ?
        acc_input_.left_cnt : step_phi;
    }
    phi_ = (phi_ + step_phi) % FULL_ANGLE;
  }
  if (acc_input_.right_cnt) {
    if (!waiting_) {
      step_phi =
        acc_input_.right_cnt < step_phi ?
        acc_input_.right_cnt : step_phi;
    }
    phi_ = (phi_ + FULL_ANGLE - step_phi) % FULL_ANGLE;
  }
}

void Game::SetWaitingXY()
{
  const Dot& d = dots_[current_dot_];
  //x_ = d.x + DoubleToInt(d.r * cos(2 * PI * phi_ / FULL_ANGLE));
  //y_ = d.y + DoubleToInt(d.r * sin(2 * PI * phi_ / FULL_ANGLE));
  x_ = d.x + RTimesCosPhi(d.r, phi_);
  y_ = d.y + RTimesSinPhi(d.r, phi_);
  
  //cout << "r: " << d.r << ", phi: " << phi_ << endl;
  //cout << "cos: " << DoubleToInt(d.r * cos(2 * PI * phi_ / FULL_ANGLE))
  //     << " : " << RTimesCosPhi(d.r, phi_) << endl;
  //cout << "sin: " << DoubleToInt(d.r * sin(2 * PI * phi_ / FULL_ANGLE))
  //     << " : " << RTimesSinPhi(d.r, phi_) << endl;
}

void Game::UpdateAccumulatedInput(const Input& input)
{
  if (input.left) {
    acc_input_.left_cnt++;
  } else {
    acc_input_.left_cnt = 0;
  }
  if (input.right) {
    acc_input_.right_cnt++;
  } else {
    acc_input_.right_cnt = 0;
  }
  if (input.accelerate) {
    acc_input_.accelerate_cnt++;
  } else {
    acc_input_.accelerate_cnt = 0;
  }
  if (input.deccelerate) {
    acc_input_.deccelerate_cnt++;
  } else {
    acc_input_.deccelerate_cnt = 0;
  }
  if (input.start) {
    acc_input_.start_cnt++;
  } else {
    acc_input_.start_cnt = 0;
  }
}

void Game::ProcessInput(const Input& input, vector<Action>* actions)
{
  UpdateAccumulatedInput(input);
  int old_phi = phi_;
  AdjustSpeedAndPhi();
  if (waiting_) {
    if (phi_ != old_phi) {
      SetWaitingXY();
    }
    if (input.start) {
      waiting_ = false;
      lines_.SetStart(x_, y_);
      AddActionHideStarter(actions);
    } else {
      if (phi_ != old_phi) {
        AddActionShowStarter(actions, current_dot_, phi_);
      }
    }
  }
}

void Game::CalculateTargetXY(int dt, int* target_x, int* target_y) const
{
  int step = speed_ * dt;
  //*target_x = x_ + DoubleToInt(step * cos(2 * PI * phi_ / FULL_ANGLE));
  //*target_y = y_ + DoubleToInt(step * sin(2 * PI * phi_ / FULL_ANGLE));
  *target_x = x_ + RTimesCosPhi(step, phi_);
  *target_y = y_ + RTimesSinPhi(step, phi_);
  
  //cout << "r: " << step << ", phi: " << phi_ << endl;
  //cout << "cos: " << DoubleToInt(step * cos(2 * PI * phi_ / FULL_ANGLE))
  //     << " : " << RTimesCosPhi(step, phi_) << endl;
  //cout << "sin: " << DoubleToInt(step * sin(2 * PI * phi_ / FULL_ANGLE))
  //     << " : " << RTimesSinPhi(step, phi_) << endl;
}

bool Game::DotHit(int from_x, int from_y, int to_x, int to_y, bool starting,
                  int* hit_x, int* hit_y, int* hit_dot_index) const
{
  int closest_x, closest_y, closest_index, min_distsq;
  bool found = false;
  for (int i = 0; i < dots_.size(); i++) {
    if (starting  &&  i == current_dot_) {
      continue;
    }
    vector<Point> isec;
    FindIntersections(from_x, from_y, to_x, to_y,
                      dots_[i].x, dots_[i].y, dots_[i].r,
                      &isec);
    for (int j = 0; j < isec.size(); j++) {
      int x = isec[j].x;
      int y = isec[j].y;
      int distsq = (x - from_x) * (x - from_x) + (y - from_y) * (y - from_y);
      if (!found || distsq < min_distsq) {
        found = true;
        min_distsq = distsq;
        closest_x = x;
        closest_y = y;
        closest_index = i;
      }
    }
  }
  if (found) {
    *hit_x = closest_x;
    *hit_y = closest_y;
    *hit_dot_index = closest_index;
  }
  return found;
}

void Game::CheckDotHitOrIntersection(int from_x, int from_y,
                                     int* to_x, int* to_y,
                                     bool starting,
                                     bool* intersection_ptr,
                                     bool* hit_ptr,
                                     int* hit_dot_index_ptr) const
{
  //cout << "Game::CheckDotHitOrIntersection(" << from_x << ", " << from_y
  //     << ", " << *to_x << ", " << *to_y << "....)\n";
  bool intersection_loc, hit_loc;
  int hit_dot_index_loc;
  bool* intersection = intersection_ptr ? intersection_ptr : &intersection_loc;
  bool* hit = hit_ptr ? hit_ptr : &hit_loc;
  int* hit_dot_index =
    hit_dot_index_ptr ? hit_dot_index_ptr : &hit_dot_index_loc;

  int intersection_x, intersection_y;
  *intersection = lines_.WouldIntersect(from_x, from_y, *to_x, *to_y,
                                        &intersection_x, &intersection_y,
                                        starting);
  if (*intersection) {
    *to_x = intersection_x;
    *to_y = intersection_y;
    if (*to_x == from_x  &&  *to_y == from_y) {
      return;
    }
  }

  int hit_x, hit_y;
  *hit = DotHit(from_x, from_y, *to_x, *to_y, starting,
                &hit_x, &hit_y, hit_dot_index);
  if (*hit) {
    *to_x = hit_x;
    *to_y = hit_y;
    *intersection = false;
  }
}

void Game::Move(int dt, bool starting, vector<Action>* actions)
{
  if (min_speed_used_ > speed_) {
    min_speed_used_ = speed_;
  }
  int target_x, target_y;
  CalculateTargetXY(dt, &target_x, &target_y);
  bool intersection;
  bool hit;
  int hit_dot_index;
  CheckDotHitOrIntersection(x_, y_, &target_x, &target_y, starting,
                            &intersection, &hit, &hit_dot_index);
  lines_.LineTo(target_x, target_y);
  AddActionDraw(actions, x_, y_, target_x, target_y);
  x_ = target_x;
  y_ = target_y;
  if (intersection ||
      (hit && dots_[hit_dot_index].type == REGULAR &&
       dots_[hit_dot_index].data.regular.nr !=
       dots_[last_regular_dot_].data.regular.nr + 1) ||
      (hit && dots_[hit_dot_index].type == EXTRA_LIFE &&
       dots_[hit_dot_index].data.extra_life.used)) {
    lives_--;
    AddActionExplosion(actions, x_, y_);
  } else if (hit && dots_[hit_dot_index].type == REGULAR) {
    current_dot_ = hit_dot_index;
    last_regular_dot_ = current_dot_;
    score_ += dots_[current_dot_].data.regular.nr;
    dots_[current_dot_].lit = true;
    AddActionLightDot(actions, current_dot_);
    waiting_ = true;
    SetWaitingXY();
    if (dots_[current_dot_].data.regular.last) {
      AddActionEndLevel(actions);
    } else {
      AddActionShowStarter(actions, current_dot_, phi_);
    }
  } else if (hit && dots_[hit_dot_index].type == EXTRA_LIFE
             && !dots_[hit_dot_index].data.extra_life.used) {
    current_dot_ = hit_dot_index;
    AddActionLightDot(actions, current_dot_);
    lives_++;
    dots_[hit_dot_index].data.extra_life.used = true;
    dots_[hit_dot_index].lit = true;
    waiting_ = true;
    SetWaitingXY();
    AddActionShowStarter(actions, current_dot_, phi_);
  } else {
    assert(!hit);
  }
}

void Game::Advance(const Input& input, int timestamp, vector<Action>* actions)
{
  //cout << "Game::Advance()\n";
  int dt = timestamp - current_timestamp_;
  assert(dt > 0);
  time_left_ -= dt;
  if (time_left_ < 0) {
    lives_--;
    AddActionOutOfTime(actions);
    return;
  }
  current_timestamp_ = timestamp;
  bool was_waiting = waiting_;
  ProcessInput(input, actions);
  if (waiting_) {
    return;
  }
  Move(dt, was_waiting, actions);
}

void Game::SetLevel(const Level& level, int level_num)
{
  lines_ = level.lines;
  dots_ = level.dots;
  phi_ = level.starting_phi % FULL_ANGLE;
  min_speed_used_ = MAX_SPEED;
  level_multiplier_ = level.multiplier;
  for (int i = 0; i < dots_.size(); i++) {
    if (dots_[i].type == REGULAR) {
      last_regular_dot_ = current_dot_ = i;
      break;
    }
  }
  current_timestamp_ = 0;
  waiting_ = true;
  time_left_ = level.max_time;
  level_num_ = level_num;
  SetWaitingXY();
}

void Game::GetDot(int dot_index, Dot* dot) const
{
  *dot = dots_[dot_index];
}

int Game::GetX() const
{
  return x_;
}

int Game::GetY() const
{
  return y_;
}

int Game::GetPhi() const
{
  return phi_;
}

int Game::GetLevelNum() const
{
  return level_num_;
}

int Game::GetScore() const
{
  return score_;
}

int Game::GetLives() const
{
  return lives_;
}

int Game::GetTimeLeft() const
{
  return time_left_;
}

int Game::GetSpeed() const
{
  return speed_;
}

int Game::GetCurrentTimestamp() const
{
  return current_timestamp_;
}

bool Game::StarterShown() const
{
  return waiting_;
}

bool Game::Initialized() const
{
  return !dots_.empty();
}

bool Game::AtLevelStart() const
{
  return current_timestamp_ == 0;
}

void Game::AddScore(int sc)
{
  score_ += sc;
}

int Game::GetMinimumSpeedUsed() const
{
  return min_speed_used_;
}

int Game::GetLevelMultiplier() const
{
  return level_multiplier_;
}

void DebugPrintAction(const Action& a)
{
  switch (a.type) {
  case DRAW:
    cout << "DRAW { start_x: " << a.data.draw.start_x
         << ", start_y: " << a.data.draw.start_y
         << ", end_x: " << a.data.draw.end_x
         << ", end_y: " << a.data.draw.end_y << " }\n";
    break;
  case END_LEVEL:
    cout << "END_LEVEL\n";
    break;
  case EXPLOSION:
    cout << "EXPLOSION { x: " << a.data.explosion.x
         << ", y: " << a.data.explosion.y << " }\n";
    break;
  case HIDE_STARTER:
    cout << "HIDE_STARTER\n";
    break;
  case LIGHT_DOT:
    cout << "LIGHT_DOT { dot_index: " << a.data.dot.dot_index << " }\n";
    break;
  case SHOW_STARTER:
    cout << "SHOW_STARTER: { dot_index: " << a.data.starter.dot_index
         << ", phi: " << a.data.starter.phi << " }\n";
    break;
  case OUT_OF_TIME:
    cout << "OUT_OF_TIME\n";
    break;
  default:
    break;
  }
}

void DebugPrintInput(const Input& input)
{
  cout << "{ ";
  if (input.left) {
    cout << "left, ";
  }
  if (input.right) {
    cout << "right, ";
  }
  if (input.accelerate) {
    cout << "accelerate, ";
  }
  if (input.deccelerate) {
    cout << "deccelerate, ";
  }
  if (input.start) {
    cout << "start, ";  
  }
  cout << " }\n";
}

}  // namespace kartofel
