//  This file is part of the computer game Kartofel.
//  Copyright (C) 2008 Pawel Aleksander Fedorynski <pfedor@fuw.edu.pl>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef KARTOFEL_GAME_H
#define KARTOFEL_GAME_H

#include <vector>
#include <istream>

#include "lines.h"

namespace kartofel {

using std::vector;
using std::istream;

enum DotType {
  REGULAR,
  EXTRA_LIFE,
};

struct Dot {
  int x;
  int y;
  int r;
  DotType type;
  bool lit;
  union {
    struct DotRegular {
      int nr;
      bool last;
    } regular;

    struct DotExtraLife {
      bool used;
    } extra_life;
  } data;
};

struct Level {
  Lines lines;
  vector<Dot> dots;
  int starting_phi;
  int max_time;
  int multiplier;
};

void LoadLevel(const string& filename, Level* level);

struct Input {
  bool left;
  bool right;
  bool accelerate;
  bool deccelerate;
  bool start;
};

// Since how many cycles have the inputs been on.  E.g., if it's the third
// time "left" is set, then left_cnt will be 3.
struct AccumulatedInput {
  AccumulatedInput()
    : left_cnt(0), right_cnt(0),
      accelerate_cnt(0), deccelerate_cnt(0),
      start_cnt(0) { }

  int left_cnt;
  int right_cnt;
  int accelerate_cnt;
  int deccelerate_cnt;
  int start_cnt;
};

enum ActionType {
  DRAW,
  END_LEVEL,
  EXPLOSION,
  HIDE_STARTER,
  LIGHT_DOT,
  SHOW_STARTER,
  OUT_OF_TIME,
};

enum {
  STEP_PHI = 15,
  STEP_PHI_STARTER = 5,
  STEP_SPEED = 1,
  MAX_SPEED = 20,
  MIN_SPEED = 2,
  LOGICAL_BOARD_WIDTH = 10000,
  LOGICAL_BOARD_HEIGHT = 10000,
  DEFAULT_MAX_TIME = 100000,
  STARTING_LIVES = 5,
  STEP_MILLISECONDS = 30,
  STEP_TS = 10,
};

struct Action {
  ActionType type;
  union {
    struct ActionDraw {
      int start_x;
      int start_y;
      int end_x;
      int end_y;
    } draw;

    struct ActionExplosion {
      int x;
      int y;
    } explosion;

    struct ActionLightDot {
      int dot_index;
    } dot;

    struct ActionShowStarter {
      int dot_index;
      int phi;
    } starter;
  } data;
};

class Game {
public:
  Game();
  void Advance(const Input& input, int timestamp, vector<Action>* actions);
  void SetLevel(const Level& level, int level_num);
  void GetDot(int dot_index, Dot* dot) const;
  int GetX() const;
  int GetY() const;
  int GetPhi() const;
  int GetLevelNum() const;
  int GetScore() const;
  int GetLives() const;
  int GetTimeLeft() const;
  int GetSpeed() const;
  int GetCurrentTimestamp() const;
  bool StarterShown() const;
  bool Initialized() const;
  void CheckDotHitOrIntersection(int from_x, int from_y,
                                 int* to_x, int* to_y,
                                 bool starting,
                                 bool* intersection,
                                 bool* hit, int* hit_dot_index) const;
  bool AtLevelStart() const;
  void AddScore(int sc);
  int GetMinimumSpeedUsed() const;
  int GetLevelMultiplier() const;
private:
  void AdjustSpeedAndPhi();
  void SetWaitingXY();
  void ProcessInput(const Input& input, vector<Action>* actions);
  void CalculateTargetXY(int dt, int* target_x, int* target_y) const;
  bool DotHit(int from_x, int from_y, int to_x, int to_y, bool starting,
              int* hit_x, int* hit_y, int* hit_dot_index) const;
  void Move(int dt, bool starting, vector<Action>* actions);
  void AddActionHideStarter(vector<Action>* actions);
  void AddActionShowStarter(vector<Action>* actions, int dot_index, int phi);
  void AddActionDraw(vector<Action>* actions, int start_x, int start_y,
                     int end_x, int end_y);
  void AddActionExplosion(vector<Action>* actions, int x, int y);
  void AddActionLightDot(vector<Action>* actions, int dot_index);
  void AddActionEndLevel(vector<Action>* actions);
  void AddActionOutOfTime(vector<Action>* actions);
  void UpdateAccumulatedInput(const Input& input);

  Lines lines_;
  vector<Dot> dots_;
  int current_dot_;
  int last_regular_dot_;
  bool waiting_;
  int current_timestamp_;
  int time_left_;
  int score_;
  int lives_;
  int x_;
  int y_;
  int phi_;
  int speed_;
  int min_speed_used_;
  int level_num_;
  int level_multiplier_;
  AccumulatedInput acc_input_;
};

void DebugPrintAction(const Action& a);
void DebugPrintInput(const Input& input);

}  // namespace kartofel

#endif  // KARTOFEL_GAME_H
