//  This file is part of the computer game Kartofel.
//  Copyright (C) 2008 Pawel Aleksander Fedorynski <pfedor@fuw.edu.pl>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <algorithm>
#include <cassert>
#include <cerrno>
#include <cstdio>
#include <curl/curl.h>
#include <iomanip>
#include <iostream>
#include <sstream>
#include "SDL.h"
#include "SDL_gfxPrimitives.h"
#include "SDL_framerate.h"
#include "SDL_image.h"
#include "SDL_mixer.h"
#include "SDL_thread.h"
#include "SDL_ttf.h"

#include "config.h"
#include "game.h"
#include "saved_game.h"
#include "submit_game.h"
#include "util.h"

using namespace std;

namespace kartofel {

enum {
  NUM_LEVELS = 21
};

enum {
    SCREEN_WIDTH = 1024,
    SCREEN_HEIGHT = 768,
    BOARD_X = 266,
    BOARD_Y = 10,
    BOARD_WIDTH = 748,
    BOARD_HEIGHT = 748,
    INFO_PANEL_X = 10,
    INFO_PANEL_Y = 284,
    INFO_PANEL_WIDTH = 246,
    INFO_PANEL_HEIGHT = 200,
    INFO_PANEL_BORDER_WIDTH = 2,
    INFO_LEVEL_X = 15,
    INFO_LEVEL_Y = 12,
    INFO_SCORE_X = 15,
    INFO_SCORE_Y = 38,
    INFO_TIME_LEFT_X = 15,
    INFO_TIME_LEFT_Y = 64,
    INFO_SPEED_X = 15,
    INFO_SPEED_Y = 90,
    INFO_LIVES_X = 15,
    INFO_LIVES_Y = 116,
    MSG_BORDER_WIDTH = 2,
    OUT_OF_TIME_MSG_BG_X = 330,
    OUT_OF_TIME_MSG_BG_Y = 300,
    OUT_OF_TIME_MSG_BG_WIDTH = 620,
    OUT_OF_TIME_MSG_BG_HEIGHT = 150,
    OUT_OF_TIME_MSG_X = 30,
    OUT_OF_TIME_MSG_Y = 30,
    GAME_OVER_MSG_BG_X = 330,
    GAME_OVER_MSG_BG_Y = 300,
    GAME_OVER_MSG_BG_WIDTH = 620,
    GAME_OVER_MSG_BG_HEIGHT = 150,
    GAME_OVER_MSG_X = 60,
    GAME_OVER_MSG_Y = 30,
    GAME_SUCCESS_MSG_BG_X = 70,
    GAME_SUCCESS_MSG_BG_Y = 200,
    GAME_SUCCESS_MSG_BG_WIDTH = 910,
    GAME_SUCCESS_MSG_BG_HEIGHT = 400,
    GAME_SUCCESS_MSG_X = 30,
    GAME_SUCCESS_MSG_Y = 30,
    END_LEVEL_BONUS_MSG_BG_X = 330,
    END_LEVEL_BONUS_MSG_BG_Y = 285,
    END_LEVEL_BONUS_MSG_BG_WIDTH = 620,
    END_LEVEL_BONUS_MSG_BG_HEIGHT = 180,
    END_LEVEL_BONUS_MSG_X = 100,
    END_LEVEL_BONUS_MSG_LEVEL_MULTIPLIER_Y = 40,
    END_LEVEL_BONUS_MSG_TIME_LEFT_Y = 70,
    END_LEVEL_BONUS_MSG_MINIMUM_SPEED_USED_Y = 100,
    END_LEVEL_BONUS_MSG_BONUS_Y = 130,
    HIGH_SCORES_BG_X = 40,
    HIGH_SCORES_BG_Y = 40,
    HIGH_SCORES_BG_WIDTH = 944,
    HIGH_SCORES_BG_HEIGHT = 688,
    HIGH_SCORES_TITLE_X = 200,
    HIGH_SCORES_TITLE_Y = 40,
    HIGH_SCORES_TEXT_X = 140,
    HIGH_SCORES_TEXT_Y = 200,
    DEFINE_KEYS_BG_X = 40,
    DEFINE_KEYS_BG_Y = 40,
    DEFINE_KEYS_BG_WIDTH = 944,
    DEFINE_KEYS_BG_HEIGHT = 688,
    DEFINE_KEYS_TITLE_X = 200,
    DEFINE_KEYS_TITLE_Y = 40,
    DEFINE_KEYS_PROMPT_X = 200,
    DEFINE_KEYS_KEY_X = 744,
    DEFINE_KEYS_LEFT_Y = 250,
    DEFINE_KEYS_RIGHT_Y = 290,
    DEFINE_KEYS_ACCELERATE_Y = 330,
    DEFINE_KEYS_DECCELERATE_Y = 370,
    DEFINE_KEYS_START_Y = 410,
    DEFINE_KEYS_FOOTER_X = 260,
    DEFINE_KEYS_FOOTER_Y = 620,
    ENTER_NAME_BG_X = 200,
    ENTER_NAME_BG_Y = 285,
    ENTER_NAME_BG_WIDTH = 620,
    ENTER_NAME_BG_HEIGHT = 180,
    ENTER_NAME_PROMPT_X = 30,
    ENTER_NAME_PROMPT_Y = 50,
    ENTER_NAME_X = 30,
    ENTER_NAME_Y = 80,
    ENTER_NAME_NOTE_X = 30,
    ENTER_NAME_NOTE_Y = 140,
    FLASHING_HALF_PERIOD = 18,
    MAX_NAME_WIDTH = 566,
    STARTER_LENGTH = 40,
    SCRATCH_WIDTH = 100,
    SCRATCH_HEIGHT = 100,
    NUM_HIGH_SCORES = 10,
};

enum Color {
  CHROME,
  BOARD,
  INFO_PANEL,
  INFO_PANEL_BORDER,
  INFO_PANEL_TEXT,
  MSG_BORDER,
  OUT_OF_TIME_MSG_BACKGROUND,
  OUT_OF_TIME_MSG,
  GAME_OVER_MSG_BACKGROUND,
  GAME_OVER_MSG,
  GAME_SUCCESS_MSG_BACKGROUND,
  GAME_SUCCESS_MSG,
  END_LEVEL_BONUS_MSG_BACKGROUND,
  END_LEVEL_BONUS_MSG,
  HIGH_SCORES_BACKGROUND,
  HIGH_SCORES_TITLE,
  HIGH_SCORES_TEXT,
  DEFINE_KEYS_BACKGROUND,
  DEFINE_KEYS_TITLE,
  DEFINE_KEYS_TEXT,
  ENTER_NAME_BACKGROUND,
  ENTER_NAME_PROMPT,
  ENTER_NAME,
  ENTER_NAME_NOTE,
  DOT_LIT,
  DOT,
  DOT_BORDER,
  DOT_LABEL,
  DOT_EXTRA_LIFE_PICTURE,
  LINE,
  MENU_BACKGROUND,
  MENU_ITEM,
  SELECTED_MENU_ITEM,
  MENU_TITLE,
  COPYRIGHT,
};

struct ColorDefinitionRGB {
  Color c;
  int r;
  int g;
  int b;
};

ColorDefinitionRGB rgb[] = {
  { CHROME, 0, 194, 197, },
  { BOARD, 197, 194, 197, },
  { INFO_PANEL, 197, 194, 197, },
  { INFO_PANEL_BORDER, 0, 0, 0, },
  { INFO_PANEL_TEXT, 0, 0, 0, },
  { MSG_BORDER, 0, 0, 0, },
  { OUT_OF_TIME_MSG_BACKGROUND, 200, 100, 100, },
  { OUT_OF_TIME_MSG, 0, 0, 0, },
  { GAME_OVER_MSG_BACKGROUND, 130, 194, 250, },
  { GAME_OVER_MSG, 0, 0, 0, },
  { GAME_SUCCESS_MSG_BACKGROUND, 250, 210, 250, },
  { GAME_SUCCESS_MSG, 0, 0, 0, },
  { END_LEVEL_BONUS_MSG_BACKGROUND, 130, 194, 250, },
  { END_LEVEL_BONUS_MSG, 0, 0, 0, },
  { HIGH_SCORES_BACKGROUND, 130, 194, 250, },
  { HIGH_SCORES_TITLE, 0, 0, 0, },
  { HIGH_SCORES_TEXT, 0, 0, 0, },
  { DEFINE_KEYS_BACKGROUND, 130, 194, 250, },
  { DEFINE_KEYS_TITLE, 0, 0, 0, },
  { DEFINE_KEYS_TEXT, 0, 0, 0, },
  { ENTER_NAME_BACKGROUND, 130, 194, 250, },
  { ENTER_NAME_PROMPT, 0, 0, 0, },
  { ENTER_NAME, 0, 0, 0, },
  { ENTER_NAME_NOTE, 0, 0, 0, },
  { DOT_LIT, 255, 255, 255, },
  { DOT, 197, 194, 0, },
  { DOT_BORDER, 0, 0, 0, },
  { DOT_LABEL, 0, 0, 0, },
  { DOT_EXTRA_LIFE_PICTURE, 0, 0, 0, },
  { LINE, 0, 0, 0, },
  { MENU_BACKGROUND, 0, 194, 197, },
  { MENU_ITEM, 255, 255, 255, },
  { SELECTED_MENU_ITEM, 0, 0, 255, },
  { MENU_TITLE, 255, 255, 255, },
  { COPYRIGHT, 210, 210, 200, },
};

enum SoundEffects {
  SOUND_MENU_NAV,
  SOUND_MENU_CLICK,
  SOUND_ROTATE,
  SOUND_START,
  SOUND_CRASH,
  SOUND_DOT_LIT,
  SOUND_LEVEL_SUCCESS,
  SOUND_OUT_OF_TIME,
  SOUND_ENTER_NAME_RETURN,
};

struct SoundEffectData {
  SoundEffects id;
  string filename;
  int volume;
  Mix_Chunk* sound;
};

SoundEffectData sound_effects[] = {
  { SOUND_MENU_NAV, SOUND_MENU_NAV_FILENAME, 100, },
  { SOUND_MENU_CLICK, SOUND_MENU_CLICK_FILENAME, 100, },
  { SOUND_ROTATE, SOUND_ROTATE_FILENAME, 20, },
  { SOUND_START, SOUND_START_FILENAME, 128, },
  { SOUND_CRASH, SOUND_CRASH_FILENAME, 128, },
  { SOUND_DOT_LIT, SOUND_DOT_LIT_FILENAME, 30, },
  { SOUND_LEVEL_SUCCESS, SOUND_LEVEL_SUCCESS_FILENAME, 128, },
  { SOUND_OUT_OF_TIME, SOUND_OUT_OF_TIME_FILENAME, 128, },
  { SOUND_ENTER_NAME_RETURN, SOUND_ENTER_NAME_RETURN_FILENAME, 100, },
};

class HighScoreTable {
public:
  HighScoreTable();
  // Returns true if "score" was high enough to be worthy of adding
  // to the table.
  bool IsHighEnough(int score);
  void Add(const string& name, int score);
  void Show(SDL_Surface* screen);
  void Get(int nr, string* name, int* score);

private:
  vector<string> names;
  vector<int> scores;
};

Level levels[NUM_LEVELS];

TTF_Font* menu_font;
TTF_Font* menu_title_font;
TTF_Font* info_font;
TTF_Font* info_font_small;
TTF_Font* msg_font;
Mix_Music* music;
SDL_Surface* menu_background_image;
SDL_Surface* physical_screen;
SDL_Surface* game_screen;
SDL_Surface* menu_screen;
SDL_Surface** current_screen;  // Points to either game_screen or menu_screen.
SDL_Surface* scratch;
int scratch_x1;
int scratch_y1;
int scratch_x2;
int scratch_y2;
bool scratch_in_use = false;

FPSmanager fps_manager;

bool fullscreen_on = false;
bool music_on = true;
bool sound_effects_on = true;
bool publish_scores = true;

HighScoreTable high_scores;
int max_starting_level = 1;

SavedGame saved_game;
bool libcurl_successfully_initialized = false;

int key_accelerate = SDLK_UP;
int key_deccelerate = SDLK_DOWN;
int key_left = SDLK_LEFT;
int key_right = SDLK_RIGHT;
int key_start = SDLK_SPACE;

void LoadHighScoresAndMaxStartingLevel()
{
  const char* filename = HIGH_SCORES_FILENAME;
  FILE* f = fopen(filename, "r");
  if (!f) {
    cerr << filename << ": " << strerror(errno) << endl;
    return;
  }
  vector<string> lines;
  char line[10000];
  // Two lines per high score (the name and the score) plus one line
  // for max starting level.
  int expected_num_lines = 2 * NUM_HIGH_SCORES + 1;
  for (int i = 0; i < expected_num_lines; i++) {
    errno = 0;
    if (!fgets(line, sizeof(line), f)) {
      if (ferror(f)) {
        cerr << filename << ": " << strerror(errno) << endl;
      } else {
        cerr << filename << ": Unexpected end of file.\n";
      }
      return;
    }
    int len = strlen(line);
    if (len > 0 || line[len - 1] == '\n') {
      line[len - 1] = '\0';
    }
    lines.push_back(line);
  }
  if (0 != fclose(f)) {
    cerr << filename << ": " << strerror(errno) << endl;
  }
  assert(lines.size() == expected_num_lines);
  max_starting_level = atoi(lines[0].c_str());
  if (max_starting_level < 1) {
    max_starting_level = 1;
  }
  if (max_starting_level > NUM_LEVELS) {
    max_starting_level = NUM_LEVELS;
  }
  for (int i = 0; i < NUM_HIGH_SCORES; i++) {
    high_scores.Add(lines[2 * i + 1], atoi(lines[2 * i + 2].c_str()));
  }
}

void SaveHighScoresAndMaxStartingLevel()
{
  const char* filename = HIGH_SCORES_FILENAME;
  FILE* f = fopen(filename, "w");
  if (!f) {
    cerr << filename << ": " << strerror(errno) << endl;
    return;
  }
  if (0 > fprintf(f, "%d\n", max_starting_level)) {
    cerr << filename << ": " << strerror(errno) << endl;
    return;
  }
  for (int i = 0; i < NUM_HIGH_SCORES; i++) {
    string name;
    int score;
    high_scores.Get(i, &name, &score);
    if (0 > fprintf(f, "%s\n%d\n", name.c_str(), score)) {
      cerr << filename << ": " << strerror(errno) << endl;
      return;
    }
  }
  if (0 != fclose(f)) {
    cerr << filename << ": " << strerror(errno) << endl;
  }
}

void Exit()
{
  SaveHighScoresAndMaxStartingLevel();
  exit(0);
}

Uint32 GetColor(Color c, SDL_Surface* s = 0)
{
  if (!s) {
    s = physical_screen;
  }
  int i;
  for (i = 0; i < sizeof(rgb)/sizeof(*rgb); ++i) {
    if (rgb[i].c == c) {
      return SDL_MapRGB(s->format, rgb[i].r, rgb[i].g, rgb[i].b);
    }
  }
  assert(!"Invalid color.");
}

SDL_Color GetSDLColor(Color c)
{
  int i;
  for (i = 0; i < sizeof(rgb)/sizeof(*rgb); ++i) {
    if (rgb[i].c == c) {
      SDL_Color retv = { rgb[i].r, rgb[i].g, rgb[i].b };
      return retv;
    }
  }
  assert(!"Invalid color.");
}

Uint32 GetGfxColor(Color c)
{
  for (int i = 0; i < sizeof(rgb)/sizeof(*rgb); ++i) {
    if (rgb[i].c == c) {
      return (static_cast<Uint32>(rgb[i].r) << 24) |
        (static_cast<Uint32>(rgb[i].g) << 16) |
        (static_cast<Uint32>(rgb[i].b) << 8) | static_cast<Uint32>(255);
    }
  }
  assert(!"Invalid color.");
}

bool PrintAt(SDL_Surface* target_screen, TTF_Font* font,
             int x, int y, const string& s, Color c,
             int* w = NULL, int* h = NULL, int max_width = -1,
             bool right_adjusted = false)
{
  if (s.empty()) {
    if (w) {
      *w = 0;
    }
    if (h) {
      *h = 0;
    }
    return true;
  }
  SDL_Color color = GetSDLColor(c);
  SDL_Surface* text = TTF_RenderUTF8_Blended(font, s.c_str(), color);
  if (!text) {
    cerr << "TTF_RenderUTF8_Blended() failed: " << TTF_GetError() << endl;
    exit(-1);
  }
  if (max_width != -1) {
    if (text->w > max_width) {
      SDL_FreeSurface(text);
      return false;
    }
  }
  SDL_Rect src, dest;
  src.x = 0;
  src.y = 0;
  src.w = text->w;
  src.h = text->h;
  if (w) {
    *w = text->w;
  }
  if (h) {
    *h = text->h;
  }
  dest.x = right_adjusted ? x - text->w : x;
  dest.y = y;
  dest.w = src.w;
  dest.h = src.h;
  SDL_BlitSurface(text, &src, target_screen, &dest);
  SDL_FreeSurface(text);
  return true;
}

void PlaySound(SoundEffects id)
{
  if (!sound_effects_on) {
    return;
  }
  for (int i = 0; i < sizeof(sound_effects)/sizeof(*sound_effects); i++) {
    if (sound_effects[i].id == id) {
      int channel = Mix_PlayChannel(-1, sound_effects[i].sound, 0);
      if (channel < 0) {
        cerr << "Mix_PlayChannel() failed: " << Mix_GetError() << endl;
        exit(-1);
      }
      Mix_Volume(channel, sound_effects[i].volume);
    }
  }
}

void CopyToScratch(int x1, int y1, int x2, int y2)
{
  assert(!scratch_in_use);
  scratch_x1 = x1;
  scratch_y1 = y1;
  scratch_x2 = x2;
  scratch_y2 = y2;
  SDL_Rect src, dest;
  src.x = x1;
  src.y = y1;
  src.w = x2 - x1 + 1;
  src.h = y2 - y1 + 1;
  dest.x = 0;
  dest.y = 0;
  SDL_BlitSurface(game_screen, &src, scratch, &dest);
  scratch_in_use = true;
}

void RetrieveFromScratch()
{
  assert(scratch_in_use);
  SDL_Rect src, dest;
  src.x = 0;
  src.y = 0;
  src.w = scratch_x2 - scratch_x1 + 1;
  src.h = scratch_y2 - scratch_y1 + 1;
  dest.x = scratch_x1;
  dest.y = scratch_y1;
  SDL_BlitSurface(scratch, &src, game_screen, &dest);
  scratch_in_use = false;
}

int XLenToScr(int xlen)
{
  return DoubleToInt(1.0 * xlen * BOARD_WIDTH / LOGICAL_BOARD_WIDTH);
}

int XToScr(int x)
{
  return BOARD_X + XLenToScr(x);
}

int YLenToScr(int ylen)
{
  return DoubleToInt(1.0 * ylen * BOARD_HEIGHT / LOGICAL_BOARD_HEIGHT);
}

int YToScr(int y)
{
  return BOARD_Y + BOARD_HEIGHT - YLenToScr(y);
}

void Update()
{
  SDL_Rect src, dest;
  src.x = 0;
  src.y = 0;
  src.w = SCREEN_WIDTH;
  src.h = SCREEN_HEIGHT;
  dest.x = 0;
  dest.y = 0;
  SDL_BlitSurface(*current_screen, &src, physical_screen, &dest);
  SDL_UpdateRect(physical_screen, 0, 0, 0, 0);
}

void ClearRect(SDL_Surface* screen, Color color,
               int x, int y, int w, int h)
{
  SDL_Rect dest;
  dest.x = x;
  dest.y = y;
  dest.w = w;
  dest.h = h;
  if (-1 == SDL_FillRect(screen, &dest, GetColor(color))) {
    cerr << "SDL_FillRect() failed: " << SDL_GetError() << endl;
    exit(-1);
  }
}

void ClearInfoPanel()
{
  ClearRect(game_screen, INFO_PANEL,
            INFO_PANEL_X + INFO_PANEL_BORDER_WIDTH,
            INFO_PANEL_Y + INFO_PANEL_BORDER_WIDTH,
            INFO_PANEL_WIDTH - 2 * INFO_PANEL_BORDER_WIDTH,
            INFO_PANEL_HEIGHT - 2 * INFO_PANEL_BORDER_WIDTH);
}

void DrawChrome()
{
  ClearRect(game_screen, CHROME, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
  ClearRect(game_screen, BOARD, BOARD_X, BOARD_Y, BOARD_HEIGHT, BOARD_WIDTH);
  ClearRect(game_screen, INFO_PANEL_BORDER,
            INFO_PANEL_X, INFO_PANEL_Y, INFO_PANEL_WIDTH, INFO_PANEL_HEIGHT);
  ClearInfoPanel();
}

void SetVideoMode()
{
  physical_screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT,
                                     0, fullscreen_on ? SDL_FULLSCREEN : 0);
  if (!physical_screen) {
    cerr << "Unable to set video mode: " << SDL_GetError() << endl;
    exit(-1);
  }
  SDL_Surface* game_screen_tmp = SDL_DisplayFormat(game_screen);
  if (!game_screen_tmp) {
    cerr << "SDL_DisplayFormat() failed: " << SDL_GetError() << endl;
    exit(-1);
  }
  SDL_FreeSurface(game_screen);
  game_screen = game_screen_tmp;
  SDL_Surface* menu_screen_tmp = SDL_DisplayFormat(menu_screen);
  if (!menu_screen_tmp) {
    cerr << "SDL_DisplayFormat() failed: " << SDL_GetError() << endl;
    exit(-1);
  }
  SDL_FreeSurface(menu_screen);
  menu_screen = menu_screen_tmp;
}

void LoadSoundEffects()
{
  for (int i = 0; i < sizeof(sound_effects)/sizeof(*sound_effects); i++) {
    sound_effects[i].sound = Mix_LoadWAV(sound_effects[i].filename.c_str());
    if (sound_effects[i].sound == NULL) {
      cerr << "Unable to load file `" << sound_effects[i].filename << "': "
           << Mix_GetError() << endl;
      exit(-1);
    }
  }
}

void AutoRepeatOn()
{
  if (-1 == SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,
                                SDL_DEFAULT_REPEAT_INTERVAL)) {
    cerr << "SDL_EnableKeyRepeat() failed: " << SDL_GetError() << endl;
  }
}

void AutoRepeatOff()
{
  if (-1 == SDL_EnableKeyRepeat(0, 0)) {
    cerr << "SDL_EnableKeyRepeat() failed: " << SDL_GetError() << endl;
  }
}

void Init()
{
  CURLcode rc = curl_global_init(CURL_GLOBAL_ALL);
  if (rc != 0) {
    cerr << "curl_global_init() failed: " << curl_easy_strerror(rc) << endl;
  } else {
    libcurl_successfully_initialized = true;
  }
  if (0 != SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO)) {
    cerr << "Unable to initialize SDL: " << SDL_GetError() << endl;
    exit(-1);
  }
  atexit(SDL_Quit);
  SDL_WM_SetIcon(IMG_Load(ICON_FILENAME), NULL);
  SDL_WM_SetCaption("Kartofel", NULL);
  if (-1 == TTF_Init()) {
    cerr << "Unable to initialize SDL_ttf: " << TTF_GetError() << endl;
    exit(-1);
  }
  atexit(TTF_Quit);
  menu_font = TTF_OpenFont(MENU_FONT_FILENAME, 20);
  if (!menu_font) {
    cerr << "Unable to load font `" << MENU_FONT_FILENAME
         << "': " << TTF_GetError() << endl;
    exit(-1);
  }
  menu_title_font = TTF_OpenFont(MENU_TITLE_FONT_FILENAME, 100);
  if (!menu_title_font) {
    cerr << "Unable to load font `" << MENU_TITLE_FONT_FILENAME
         << "': " << TTF_GetError() << endl;
    exit(-1);
  }
  info_font = TTF_OpenFont(INFO_FONT_FILENAME, 20);
  if (!info_font) {
    cerr << "Unable to load font `" << INFO_FONT_FILENAME
         << "': " << TTF_GetError() << endl;
    exit(-1);
  }
  info_font_small = TTF_OpenFont(INFO_FONT_SMALL_FILENAME, 13);
  if (!info_font_small) {
    cerr << "Unable to load font `" << INFO_FONT_SMALL_FILENAME
         << "': " << TTF_GetError() << endl;
    exit(-1);
  }
  msg_font = TTF_OpenFont(MSG_FONT_FILENAME, 60);
  if (!msg_font) {
    cerr << "Unable to load font `" << MSG_FONT_FILENAME
         << "': " << TTF_GetError() << endl;
    exit(-1);
  }
  if (Mix_OpenAudio(44100, AUDIO_S16SYS, 2, 2048)) {
    cerr << "Unable to open audio: " << Mix_GetError() << endl;
    exit(-1);
  }
  atexit(Mix_CloseAudio);
  music = Mix_LoadMUS(MUSIC_FILENAME);
  if (!music) {
    cerr << "Unable to load music from file `" << MUSIC_FILENAME << "': "
         << Mix_GetError() << endl;
    exit(-1);
  }
  LoadSoundEffects();
  Mix_AllocateChannels(100);
  menu_background_image = IMG_Load(MENU_BACKGROUND_IMAGE_FILENAME);
  if (!menu_background_image) {
    cerr << "Unable to load image from file `"
         << MENU_BACKGROUND_IMAGE_FILENAME << "': " << IMG_GetError() << endl;
  }

  game_screen = SDL_CreateRGBSurface(
    SDL_SWSURFACE, SCREEN_WIDTH, SCREEN_HEIGHT, 32, 0, 0, 0, 0);
  if (!game_screen) {
    cerr << "SDL_CreateRGBSurface() failed: " << SDL_GetError() << endl;
    exit(-1);
  }
  menu_screen = SDL_CreateRGBSurface(
    SDL_SWSURFACE, SCREEN_WIDTH, SCREEN_HEIGHT, 32, 0, 0, 0, 0);
  if (!menu_screen) {
    cerr << "SDL_CreateRGBSurface() failed: " << SDL_GetError() << endl;
    exit(-1);
  }
  scratch = SDL_CreateRGBSurface(
    SDL_SWSURFACE, SCRATCH_WIDTH, SCRATCH_HEIGHT, 32, 0, 0, 0, 0);
  if (!scratch) {
    cerr << "SDL_CreateRGBSurface() failed: " << SDL_GetError() << endl;
    exit(-1);
  }
  SetVideoMode();
  current_screen = &menu_screen;
  AutoRepeatOff();
  LoadHighScoresAndMaxStartingLevel();
  SDL_initFramerate(&fps_manager);
  if (-1 == SDL_setFramerate(&fps_manager, 1000 / STEP_MILLISECONDS)) {
    cerr << "SDL_setFramerate() failed.\n";
    exit(-1);
  }
}

void DrawDot(const Dot& dot, bool lit)
{
  int xscr = XToScr(dot.x);
  int yscr = YToScr(dot.y);
  int rxscr = XLenToScr(dot.r);
  int ryscr = YLenToScr(dot.r);
  Uint32 color = GetGfxColor(lit ? DOT_LIT : DOT);
  int rc = filledEllipseColor(game_screen, xscr, yscr, rxscr, ryscr, color);
  assert(rc == 0);
  rc = aaellipseColor(game_screen, xscr, yscr, rxscr - 1, ryscr - 1,
                      GetGfxColor(DOT_BORDER));
  rc |= aaellipseColor(game_screen, xscr, yscr, rxscr, ryscr,
                       GetGfxColor(DOT_BORDER));
  assert(rc == 0);

  if (dot.type == REGULAR) {
    int dot_nr = dot.data.regular.nr;
    stringstream ss;
    ss << dot_nr;
    int x = dot_nr >= 10 ? xscr - 8 : xscr - 3;
    int y = yscr - 3;
    rc = stringColor(game_screen, x, y, ss.str().c_str(),
                     GetGfxColor(DOT_LABEL));
    assert(rc == 0);
  } else if (dot.type == EXTRA_LIFE) {
    int u = dot.r / 12;

    // Head.
    int xhead = XToScr(dot.x);
    int yhead = YToScr(dot.y + 6 * u);
    int rxhead = XLenToScr(2 * u);
    int ryhead = YLenToScr(2 * u);

    // Hands.
    int xlhf = XToScr(dot.x - 5 * u);
    int ylhf = YToScr(dot.y);
    int xlht = XToScr(dot.x);
    int ylht = YToScr(dot.y + 3 * u);

    int xrhf = XToScr(dot.x + 5 * u);
    int yrhf = YToScr(dot.y);
    int xrht = XToScr(dot.x);
    int yrht = YToScr(dot.y + 3 * u);

    // Neck and torso.
    int xtf = XToScr(dot.x);
    int ytf = YToScr(dot.y + 4 * u);
    int xtt = XToScr(dot.x);
    int ytt = YToScr(dot.y - u);

    // Legs.
    int xllf = XToScr(dot.x - 4 * u);
    int yllf = YToScr(dot.y - 8 * u);
    int xllt = XToScr(dot.x);
    int yllt = YToScr(dot.y - u);

    int xrlf = XToScr(dot.x + 4 * u);
    int yrlf = YToScr(dot.y - 8 * u);
    int xrlt = XToScr(dot.x);
    int yrlt = YToScr(dot.y - u);

    rc = aaellipseColor(game_screen, xhead, yhead, rxhead, ryhead,
                        GetGfxColor(DOT_EXTRA_LIFE_PICTURE));
    assert(rc == 0);
    rc = aalineColor(game_screen, xlhf, ylhf, xlht, ylht,
                     GetGfxColor(DOT_EXTRA_LIFE_PICTURE));
    assert(rc == 0);
    rc = aalineColor(game_screen, xrhf, yrhf, xrht, yrht,
                     GetGfxColor(DOT_EXTRA_LIFE_PICTURE));
    assert(rc == 0);
    rc = aalineColor(game_screen, xllf, yllf, xllt, yllt,
                     GetGfxColor(DOT_EXTRA_LIFE_PICTURE));
    assert(rc == 0);
    rc = aalineColor(game_screen, xrlf, yrlf, xrlt, yrlt,
                     GetGfxColor(DOT_EXTRA_LIFE_PICTURE));
    assert(rc == 0);
    rc = aalineColor(game_screen, xtf, ytf, xtt, ytt,
                     GetGfxColor(DOT_EXTRA_LIFE_PICTURE));
    assert(rc == 0);
  }
}

void DrawLine(int x1, int y1, int x2, int y2, bool thick = true)
{
  int x1scr = XToScr(x1);
  int y1scr = YToScr(y1);
  int x2scr = XToScr(x2);
  int y2scr = YToScr(y2);
  Uint32 color = GetGfxColor(LINE);
  int rc = aalineColor(game_screen, x1scr, y1scr, x2scr, y2scr, color);
  if (thick) {
    rc |= aalineColor(game_screen, x1scr + 1, y1scr, x2scr + 1, y2scr, color);
    rc |= aalineColor(game_screen, x1scr, y1scr + 1, x2scr, y2scr + 1, color);
  }
  assert(rc == 0);
}

void DrawLevel(const Level& level)
{
  DrawChrome();
  int x1, y1, x2, y2;
  level.lines.Reset();
  while (level.lines.GetNext(&x1, &y1, &x2, &y2)) {
    DrawLine(x1, y1, x2, y2);
  }
  bool first_regular_found = false;
  for (int i = 0; i < level.dots.size(); i++) {
    bool lit = false;
    if (level.dots[i].type == REGULAR && !first_regular_found) {
      lit = true;
      first_regular_found = true;
    }
    DrawDot(level.dots[i], lit);
  }
  //Update();
}

void ReadEvents(Input* input, bool* pause)
{
  //cout << "ReadEvents()\n";
  SDL_Event e;
  while (SDL_PollEvent(&e)) {
    if (e.type == SDL_QUIT) {
      cout << "Bye.\n";
      Exit();
    }
  }
  Uint8* keys = SDL_GetKeyState(NULL);
  *pause = keys[SDLK_ESCAPE];
  input->accelerate = keys[key_accelerate];
  input->deccelerate = keys[key_deccelerate];
  input->left = keys[key_left];
  input->right = keys[key_right];
  input->start = keys[key_start];
  //cout << "Input: { " << "acc: " << input->accelerate
  //     << ", decc: " << input->deccelerate
  //     << ", left: " << input->left << ", right: " << input->right
  //     << ", start: " << input->start << " }\n";
}

void HideStarter()
{
  RetrieveFromScratch();
}

void ShowStarter(const Game& game)
{
  int x1 = game.GetX();
  int y1 = game.GetY();
  int phi = game.GetPhi();
  int x2 = x1 + DoubleToInt(STARTER_LENGTH * cos(2 * PI * phi / FULL_ANGLE));
  int y2 = y1 + DoubleToInt(STARTER_LENGTH * sin(2 * PI * phi / FULL_ANGLE));
  bool is_intersection, is_dot_hit;
  game.CheckDotHitOrIntersection(x1, y1, &x2, &y2, true,
                                 &is_intersection, &is_dot_hit, NULL);
  int x1scr = XToScr(x1);
  int y1scr = YToScr(y1);
  int x2scr = XToScr(x2);
  int y2scr = YToScr(y2);
  if (x1scr > x2scr) {
    swap(x1scr, x2scr);
  }
  if (y1scr > y2scr) {
    swap(y1scr, y2scr);
  }
  CopyToScratch(x1scr, y1scr, x2scr + 1, y2scr + 1);
  DrawLine(x1, y1, x2, y2, !is_intersection && !is_dot_hit);
}

struct EndLevel {
  bool success;
  bool death;
};

void MsgOutOfTime()
{
  ClearRect(game_screen, MSG_BORDER,
            OUT_OF_TIME_MSG_BG_X, OUT_OF_TIME_MSG_BG_Y,
            OUT_OF_TIME_MSG_BG_WIDTH, OUT_OF_TIME_MSG_BG_HEIGHT);
  ClearRect(game_screen, OUT_OF_TIME_MSG_BACKGROUND,
            OUT_OF_TIME_MSG_BG_X + MSG_BORDER_WIDTH,
            OUT_OF_TIME_MSG_BG_Y + MSG_BORDER_WIDTH,
            OUT_OF_TIME_MSG_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            OUT_OF_TIME_MSG_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);
  PrintAt(game_screen, msg_font,
          OUT_OF_TIME_MSG_BG_X + OUT_OF_TIME_MSG_X,
          OUT_OF_TIME_MSG_BG_Y + OUT_OF_TIME_MSG_Y,
          "OUT OF TIME", OUT_OF_TIME_MSG);
}

void MsgGameOver()
{
  ClearRect(game_screen, MSG_BORDER,
            GAME_OVER_MSG_BG_X, GAME_OVER_MSG_BG_Y,
            GAME_OVER_MSG_BG_WIDTH, GAME_OVER_MSG_BG_HEIGHT);
  ClearRect(game_screen, GAME_OVER_MSG_BACKGROUND,
            GAME_OVER_MSG_BG_X + MSG_BORDER_WIDTH,
            GAME_OVER_MSG_BG_Y + MSG_BORDER_WIDTH,
            GAME_OVER_MSG_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            GAME_OVER_MSG_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);
  PrintAt(game_screen, msg_font,
          GAME_OVER_MSG_BG_X + GAME_OVER_MSG_X,
          GAME_OVER_MSG_BG_Y + GAME_OVER_MSG_Y,
          "GAME OVER", GAME_OVER_MSG);
}

void MsgGameSuccess()
{
  ClearRect(game_screen, MSG_BORDER,
            GAME_SUCCESS_MSG_BG_X, GAME_SUCCESS_MSG_BG_Y,
            GAME_SUCCESS_MSG_BG_WIDTH, GAME_SUCCESS_MSG_BG_HEIGHT);
  ClearRect(game_screen, GAME_SUCCESS_MSG_BACKGROUND,
            GAME_SUCCESS_MSG_BG_X + MSG_BORDER_WIDTH,
            GAME_SUCCESS_MSG_BG_Y + MSG_BORDER_WIDTH,
            GAME_SUCCESS_MSG_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            GAME_SUCCESS_MSG_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);
  PrintAt(game_screen, msg_font,
          GAME_SUCCESS_MSG_BG_X + GAME_SUCCESS_MSG_X,
          GAME_SUCCESS_MSG_BG_Y + GAME_SUCCESS_MSG_Y,
          "CONGRATULATIONS!", GAME_SUCCESS_MSG);
  PrintAt(game_screen, info_font,
          GAME_SUCCESS_MSG_BG_X + GAME_SUCCESS_MSG_X + 210,
          GAME_SUCCESS_MSG_BG_Y + GAME_SUCCESS_MSG_Y + 140,
          "You have completed all levels.", GAME_SUCCESS_MSG);
  PrintAt(game_screen, info_font,
          GAME_SUCCESS_MSG_BG_X + GAME_SUCCESS_MSG_X + 210,
          GAME_SUCCESS_MSG_BG_Y + GAME_SUCCESS_MSG_Y + 170,
          "You are the master.", GAME_SUCCESS_MSG);
  PrintAt(game_screen, info_font,
          GAME_SUCCESS_MSG_BG_X + GAME_SUCCESS_MSG_X + 210,
          GAME_SUCCESS_MSG_BG_Y + GAME_SUCCESS_MSG_Y + 200,
          "Thank you for playing Kartofel.", GAME_SUCCESS_MSG);
  PrintAt(game_screen, info_font,
          GAME_SUCCESS_MSG_BG_X + GAME_SUCCESS_MSG_X + 210,
          GAME_SUCCESS_MSG_BG_Y + GAME_SUCCESS_MSG_Y + 230,
          "Press Enter to continue.", GAME_SUCCESS_MSG);
}

void ProcessAction(const Game& game, const Action& a,
                   bool* starter_shown, EndLevel* end_level, bool* dot_lit)
{
  //DebugPrintAction(a);
  switch (a.type) {
  case DRAW:
    DrawLine(a.data.draw.start_x, a.data.draw.start_y,
             a.data.draw.end_x, a.data.draw.end_y);
    break;
  case END_LEVEL:
    end_level->success = true;
    break;
  case EXPLOSION:
    PlaySound(SOUND_CRASH);
    end_level->death = true;
    break;
  case OUT_OF_TIME:
    PlaySound(SOUND_OUT_OF_TIME);
    MsgOutOfTime();
    end_level->death = true;
    break;
  case HIDE_STARTER:
    HideStarter();
    *starter_shown = false;
    break;
  case LIGHT_DOT:
    Dot dot;
    game.GetDot(a.data.dot.dot_index, &dot);
    DrawDot(dot, true);
    *dot_lit = true;
    break;
  case SHOW_STARTER:
    if (*starter_shown) {
      HideStarter();
    }
    ShowStarter(game);
    *starter_shown = true;
    break;
  default:
    break;
  }
}

int WaitForKey(int* unicode = NULL, string* name = NULL)
{
  int old_u;
  if (unicode) {
    old_u = SDL_EnableUNICODE(1);
  }
  SDL_Event e;
  for (;;) {
    if (0 == SDL_WaitEvent(&e)) {
      cerr <<  "Error.\n";
      exit(1);
    }
    switch (e.type) {
    case SDL_QUIT:
      cout << "Bye.\n";
      Exit();
    case SDL_KEYDOWN:
      //cout << "Key pressed: " << SDL_GetKeyName(e.key.keysym.sym) << endl;
      if (unicode) {
        *unicode = e.key.keysym.unicode;
        SDL_EnableUNICODE(old_u);
      }
      if (name) {
        *name = SDL_GetKeyName(e.key.keysym.sym);
      }
      return e.key.keysym.sym;
    default:
      break;
    }
  }
}

enum PlayLevelResult {
  LEVEL_SUCCESS = 0,
  LEVEL_DEATH = 1,
  LEVEL_PAUSE = 2,
};

string FormattedTimeLeft(int time_left, int* tlms_out = NULL)
{
  int tlms = time_left * STEP_MILLISECONDS / STEP_TS;
  if (tlms_out) {
    *tlms_out = tlms;
  }
  int tlsec = tlms / 1000;
  int tlmin = tlsec / 60;
  char buf[100];
  snprintf(buf, sizeof(buf), "%02d:%02d.%d",
           tlmin, tlsec - tlmin * 60, (tlms - tlsec * 1000) / 100);
  return buf;
}

void PrintInfo(const Game& game)
{
  ClearInfoPanel();
  stringstream ss;
  ss << "Level: " << game.GetLevelNum();
  PrintAt(game_screen, info_font,
          INFO_PANEL_X + INFO_LEVEL_X,
          INFO_PANEL_Y + INFO_LEVEL_Y,
          ss.str(),
          INFO_PANEL_TEXT);
  ss.str("");
  ss << "Score: " << game.GetScore();
  PrintAt(game_screen, info_font,
          INFO_PANEL_X + INFO_SCORE_X,
          INFO_PANEL_Y + INFO_SCORE_Y,
          ss.str(),
          INFO_PANEL_TEXT);
  ss.str("");
  ss << "Time left: " << FormattedTimeLeft(game.GetTimeLeft());
  PrintAt(game_screen, info_font,
          INFO_PANEL_X + INFO_TIME_LEFT_X,
          INFO_PANEL_Y + INFO_TIME_LEFT_Y,
          ss.str(),
          INFO_PANEL_TEXT);
  ss.str("");
  ss << "Speed: " << game.GetSpeed();
  PrintAt(game_screen, info_font,
          INFO_PANEL_X + INFO_SPEED_X,
          INFO_PANEL_Y + INFO_SPEED_Y,
          ss.str(),
          INFO_PANEL_TEXT);
  ss.str("");
  ss << "Lives: " << game.GetLives();
  PrintAt(game_screen, info_font,
          INFO_PANEL_X + INFO_LIVES_X,
          INFO_PANEL_Y + INFO_LIVES_Y,
          ss.str(),
          INFO_PANEL_TEXT);
}

PlayLevelResult PlayLevel(Game* game)
{
  Input input = { false, false, false, false, false };
  int timestamp = game->GetCurrentTimestamp();
  bool starter_shown = game->StarterShown();
  PrintInfo(*game);
  Update();
  if (game->AtLevelStart()) {
    // In case someone pressed the "Go" key after finishing level, we
    // don't want to interpret the still-pressed key as input and start
    // immediately.
    WaitForKey();
  }
  int old_phi = game->GetPhi();
  for (;;) {
    bool old_starter_shown = starter_shown;
    //SDL_Delay(STEP_MILLISECONDS);
    SDL_framerateDelay(&fps_manager);
    bool pause;
    ReadEvents(&input, &pause);
    if (pause) {
      return LEVEL_PAUSE;
    }
    timestamp += STEP_TS;
    //cout << "timestamp: " << timestamp << endl;
    //DebugPrintInput(input);
    saved_game.AddStep(input);
    vector<Action> actions;
    game->Advance(input, timestamp, &actions);
    bool dot_lit = false;
    EndLevel end_level = { false, false };
    for (int i = 0; i < actions.size(); i++) {
      ProcessAction(*game, actions[i], &starter_shown, &end_level, &dot_lit);
    }
    if (old_phi != game->GetPhi() && starter_shown) {
      PlaySound(SOUND_ROTATE);
    }
    if (old_starter_shown && !starter_shown) {
      PlaySound(SOUND_START);
    }
    if (dot_lit /* && !end_level.succes */) {
      PlaySound(SOUND_DOT_LIT);
    }
    PrintInfo(*game);
    Update();
    if (end_level.success) {
      PlaySound(SOUND_LEVEL_SUCCESS);
      return LEVEL_SUCCESS;
    } else if (end_level.death) {
      return LEVEL_DEATH;
    }
    old_phi = game->GetPhi();
  }
}

enum PlayGameResult {
  GAME_SUCCESS = 0,
  GAME_DEATH = 1,
  GAME_PAUSE = 2,
};

int WaitForEnterOrEsc()
{
  for (;;) {
    int key = WaitForKey();
    if (key == SDLK_RETURN || key == SDLK_ESCAPE) {
      return key;
    }
  }
}

void InitializeLevel(Game* game, int level)
{
  // Levels in game are numbered from 1, but indices in the "levels"
  // array from 0.
  game->SetLevel(levels[level - 1], level);
  DrawLevel(levels[level - 1]);
  scratch_in_use = false;
  ShowStarter(*game);
}

bool CheckForKey(int* key)
{
  SDL_Event e;
  while (SDL_PollEvent(&e)) {
    if (e.type == SDL_QUIT) {
      cout << "Bye.\n";
      Exit();
    }
    if (e.type == SDL_KEYDOWN) {
      *key = e.key.keysym.sym;
      return true;
    }
  }
  return false;
}

int WaitForKeyWithFlashing(const Game* game)
{
  const Level& level = levels[game->GetLevelNum() - 1];
  int key = SDLK_RETURN;
  for (int i = 0; i < 10; i++) {
    for (int j = 0; j < 2; j++) {
      for (int k = 0; k < level.dots.size(); k++) {
        Dot dot;
        game->GetDot(k, &dot);
        DrawDot(level.dots[k], j == 0 && dot.lit);
      }
      Update();
      for (int k = 0; k < 10; k++) {
        if (CheckForKey(&key)) {
          goto end;
        }
        SDL_Delay(FLASHING_HALF_PERIOD);
      }
    }
  }
end:
  for (int i = 0; i < level.dots.size(); i++) {
    Dot dot;
    game->GetDot(i, &dot);
    DrawDot(level.dots[i], dot.lit);
  }
  Update();
  return key;
}

void MsgEndLevelBonus(const string& msg_level_multiplier,
                      const string& msg_time_left,
                      const string& msg_minimum_speed_used,
                      const string& msg_bonus)
{
  ClearRect(game_screen, MSG_BORDER,
            END_LEVEL_BONUS_MSG_BG_X, END_LEVEL_BONUS_MSG_BG_Y,
            END_LEVEL_BONUS_MSG_BG_WIDTH, END_LEVEL_BONUS_MSG_BG_HEIGHT);
  ClearRect(game_screen, END_LEVEL_BONUS_MSG_BACKGROUND,
            END_LEVEL_BONUS_MSG_BG_X + MSG_BORDER_WIDTH,
            END_LEVEL_BONUS_MSG_BG_Y + MSG_BORDER_WIDTH,
            END_LEVEL_BONUS_MSG_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            END_LEVEL_BONUS_MSG_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);
  PrintAt(game_screen, info_font,
          END_LEVEL_BONUS_MSG_BG_X + END_LEVEL_BONUS_MSG_X,
          END_LEVEL_BONUS_MSG_BG_Y + END_LEVEL_BONUS_MSG_LEVEL_MULTIPLIER_Y,
          msg_level_multiplier, END_LEVEL_BONUS_MSG);
  PrintAt(game_screen, info_font,
          END_LEVEL_BONUS_MSG_BG_X + END_LEVEL_BONUS_MSG_X,
          END_LEVEL_BONUS_MSG_BG_Y + END_LEVEL_BONUS_MSG_TIME_LEFT_Y,
          msg_time_left, END_LEVEL_BONUS_MSG);
  PrintAt(game_screen, info_font,
          END_LEVEL_BONUS_MSG_BG_X + END_LEVEL_BONUS_MSG_X,
          END_LEVEL_BONUS_MSG_BG_Y + END_LEVEL_BONUS_MSG_MINIMUM_SPEED_USED_Y,
          msg_minimum_speed_used, END_LEVEL_BONUS_MSG);
  PrintAt(game_screen, info_font,
          END_LEVEL_BONUS_MSG_BG_X + END_LEVEL_BONUS_MSG_X,
          END_LEVEL_BONUS_MSG_BG_Y + END_LEVEL_BONUS_MSG_BONUS_Y,
          msg_bonus, END_LEVEL_BONUS_MSG);
}

int EndLevelBonus(Game* game, bool show_msg)
{
  int min_speed = game->GetMinimumSpeedUsed();
  int tlms;
  string time_left_str = FormattedTimeLeft(game->GetTimeLeft(), &tlms);
  int tlds = tlms / 100;
  int level_multiplier = game->GetLevelMultiplier();
  int bonus = level_multiplier * tlds * min_speed * min_speed;
  game->AddScore(bonus);
  if (show_msg) {
    char msg_level_multiplier[1000];
    snprintf(msg_level_multiplier, sizeof(msg_level_multiplier),
             "Level multiplier: %d", game->GetLevelMultiplier());
    int tlsec = tlds / 10;
    char msg_time_left[1000];
    if (tlds > 600) {
      snprintf(msg_time_left, sizeof(msg_time_left),
               "Time left: %s = %d.%d seconds", time_left_str.c_str(),
               tlsec, tlds - tlsec * 10);
    } else {
      snprintf(msg_time_left, sizeof(msg_time_left),
               "Time left: %d.%d seconds", tlsec, tlds - tlsec * 10);
    }
    char msg_minimum_speed_used[1000];
    snprintf(msg_minimum_speed_used, sizeof(msg_minimum_speed_used),
             "Minimum speed used: %d", min_speed);
    char msg_bonus[1000];
    snprintf(
      msg_bonus, sizeof(msg_bonus),
      "Bonus: %d \xc3\x97 %d.%d \xc3\x97 %d \xc3\x97 %d \xc3\x97 10 = %d",
      level_multiplier, tlsec, tlds - tlsec * 10, min_speed, min_speed, bonus);
    MsgEndLevelBonus(msg_level_multiplier, msg_time_left,
                     msg_minimum_speed_used, msg_bonus);
    PrintInfo(*game);
    Update();
    return WaitForKey();
  }
  return SDLK_RETURN;
}

PlayGameResult PlayGame(Game* game, int starting_level = 1)
{
  assert(starting_level > 0 && starting_level <= NUM_LEVELS);
  if (!game->Initialized()) {
    InitializeLevel(game, starting_level);
  }
  for (;;) {
    PlayLevelResult res = PlayLevel(game);
    if (res == LEVEL_PAUSE) {
      return GAME_PAUSE;
    } else if (res == LEVEL_SUCCESS) {
      int current_level = game->GetLevelNum();
      if (current_level == NUM_LEVELS) {
        WaitForKeyWithFlashing(game);
        EndLevelBonus(game, false);
        return GAME_SUCCESS;
      } else {
        int key = WaitForKeyWithFlashing(game);
        int key2 = EndLevelBonus(game, key != SDLK_ESCAPE);
        current_level++;
        if (current_level > max_starting_level) {
          max_starting_level = current_level;
        }
        InitializeLevel(game, current_level);
        if (key == SDLK_ESCAPE || key2 == SDLK_ESCAPE) {
          return GAME_PAUSE;
        }
      }
    } else if (res == LEVEL_DEATH) {
      if (game->GetLives() == 0) {
        return GAME_DEATH;
      }
      int key = WaitForKey();
      InitializeLevel(game, game->GetLevelNum());
      if (key == SDLK_ESCAPE) {
        return GAME_PAUSE;
      }
    }
  }
}

// ====

bool DefineOneKey(const string& prompt, int y, int* k)
{
  PrintAt(menu_screen, info_font,
          DEFINE_KEYS_BG_X + DEFINE_KEYS_PROMPT_X,
          DEFINE_KEYS_BG_Y + y,
          prompt, DEFINE_KEYS_TEXT);
  Update();
  string key_name;
  int key = WaitForKey(NULL, &key_name);
  if (key == SDLK_ESCAPE) {
    return false;
  }
  *k = key;
  transform(key_name.begin(), key_name.end(), key_name.begin(), ::toupper);
  PrintAt(menu_screen, info_font,
          DEFINE_KEYS_BG_X + DEFINE_KEYS_KEY_X,
          DEFINE_KEYS_BG_Y + y,
          key_name,
          DEFINE_KEYS_TEXT,
          NULL, NULL, -1, true);
  PlaySound(SOUND_MENU_NAV);
  Update();

  return true;
}

void DefineKeys()
{
  int key_left_local, key_right_local, key_accelerate_local,
    key_deccelerate_local, key_start_local;

  ClearRect(menu_screen, MSG_BORDER,
            DEFINE_KEYS_BG_X, DEFINE_KEYS_BG_Y,
            DEFINE_KEYS_BG_WIDTH, DEFINE_KEYS_BG_HEIGHT);
  ClearRect(menu_screen, DEFINE_KEYS_BACKGROUND,
            DEFINE_KEYS_BG_X + MSG_BORDER_WIDTH,
            DEFINE_KEYS_BG_Y + MSG_BORDER_WIDTH,
            DEFINE_KEYS_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            DEFINE_KEYS_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);

  PrintAt(menu_screen, msg_font,
          DEFINE_KEYS_BG_X + DEFINE_KEYS_TITLE_X,
          DEFINE_KEYS_BG_Y + DEFINE_KEYS_TITLE_Y,
          "Define Keys", DEFINE_KEYS_TITLE);

  if (!DefineOneKey("Turn left:", DEFINE_KEYS_LEFT_Y, &key_left_local)) {
    return;
  }
  if (!DefineOneKey("Turn right:", DEFINE_KEYS_RIGHT_Y, &key_right_local)) {
    return;
  }
  if (!DefineOneKey("Speed up:", DEFINE_KEYS_ACCELERATE_Y,
                    &key_accelerate_local)) {
    return;
  }
  if (!DefineOneKey("Slow down:", DEFINE_KEYS_DECCELERATE_Y,
                    &key_deccelerate_local)) {
    return;
  }
  if (!DefineOneKey("Go:", DEFINE_KEYS_START_Y, &key_start_local)) {
    return;
  }
  PrintAt(menu_screen, info_font,
          DEFINE_KEYS_BG_X + DEFINE_KEYS_FOOTER_X,
          DEFINE_KEYS_BG_Y + DEFINE_KEYS_FOOTER_Y,
          "Press Enter to confirm or Escape to cancel.",
          DEFINE_KEYS_TEXT);
  Update();

  int key = WaitForEnterOrEsc();
  if (key == SDLK_RETURN) {
    PlaySound(SOUND_MENU_CLICK);
    key_left = key_left_local;
    key_right = key_right_local;
    key_accelerate = key_accelerate_local;
    key_deccelerate = key_deccelerate_local;
    key_start = key_start_local;
  }
}

void ToggleFullscreen()
{
  fullscreen_on = !fullscreen_on;
  SetVideoMode();
  Update();
}

void ToggleMusic()
{
  music_on = !music_on;
  if (music_on) {
    Mix_ResumeMusic();
  } else {
    Mix_PauseMusic();
  }
}

void ToggleSoundEffects()
{
  sound_effects_on = !sound_effects_on;
}

void TogglePublishScores()
{
  publish_scores = !publish_scores;
}

void ShowHelp()
{

}

HighScoreTable::HighScoreTable()
{
  for (int i = 0; i < NUM_HIGH_SCORES; i++) {
    names.push_back("..........");
    scores.push_back(0);
  }
}

void HighScoreTable::Get(int nr, string* name, int* score)
{
  assert(nr >= 0 && nr < NUM_HIGH_SCORES);
  *name = names[nr];
  *score = scores[nr];
}

bool HighScoreTable::IsHighEnough(int score)
{
  return score > scores.back();
}

void HighScoreTable::Add(const string& name, int score)
{
  int i;
  for (i = scores.size() - 1; i > 0 && scores[i - 1] < score; i--) {
    names[i] = names[i - 1];
    scores[i] = scores[i - 1];
  }
  names[i] = name;
  scores[i] = score;
}

void HighScoreTable::Show(SDL_Surface* screen)
{
  ClearRect(screen, MSG_BORDER,
            HIGH_SCORES_BG_X, HIGH_SCORES_BG_Y,
            HIGH_SCORES_BG_WIDTH, HIGH_SCORES_BG_HEIGHT);
  ClearRect(screen, HIGH_SCORES_BACKGROUND,
            HIGH_SCORES_BG_X + MSG_BORDER_WIDTH,
            HIGH_SCORES_BG_Y + MSG_BORDER_WIDTH,
            HIGH_SCORES_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            HIGH_SCORES_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);

  PrintAt(screen, msg_font,
          HIGH_SCORES_BG_X + HIGH_SCORES_TITLE_X,
          HIGH_SCORES_BG_Y + HIGH_SCORES_TITLE_Y,
          "High Scores", HIGH_SCORES_TITLE);

  int x = HIGH_SCORES_TEXT_X;
  int y = HIGH_SCORES_TEXT_Y;
  for (int i = 0; i < scores.size(); i++) {
    stringstream ss;
    ss << (i + 1) << ". ";
    PrintAt(screen, info_font,
            HIGH_SCORES_BG_X + x, HIGH_SCORES_BG_Y + y,
            ss.str(), HIGH_SCORES_TEXT, NULL, NULL, -1, true);
    PrintAt(screen, info_font,
            HIGH_SCORES_BG_X + x, HIGH_SCORES_BG_Y + y,
            names[i], HIGH_SCORES_TEXT);
    ss.str("");
    ss << scores[i];
    PrintAt(screen, info_font,
            HIGH_SCORES_BG_X + x + MAX_NAME_WIDTH + 120,
            HIGH_SCORES_BG_Y + y, ss.str(), HIGH_SCORES_TEXT,
            NULL, NULL, -1, true);
    y += 40;
  }
}

// - Start New Game (Level ...)
// [Not always] - Continue Game
// - Define Keys
// - High Scores
// - Fullscreen on/off
// - Music on/off
// - Sound Effects on/off
// - Publish Scores yes/no
// - Help
// - Quit
class Menu {
public:
  Menu();
  void EnableContinue();
  void DisableContinue();
  int GetStartingLevel() const;

  enum Choice {
    CHOICE_NEW_GAME,
    CHOICE_CONTINUE,
  };

  Choice Run();

private:
  enum Item {
    NEW_GAME = 0,
    CONTINUE_GAME = 1,
    DEFINE_KEYS = 2,
    HIGH_SCORES = 3,
    FULLSCREEN = 4,
    MUSIC = 5,
    SOUND_EFFECTS = 6,
    PUBLISH_SCORES = 7,
    HELP = 8,
    QUIT = 9,
    NUM_ITEMS = 10,
  };
  vector<string> text;
  vector<bool> active;
  int selected_item;
  vector<Point> pos;
  vector<int> item_width;
  vector<int> item_height;
  int starting_level;

  void Draw();
  void PrintItem(int item, bool selected, const char* more = NULL);
  void PrintItem(int item, Color color, const char* more = NULL);
  void EraseItem(int item);
};

Menu::Menu()
{
  text.push_back("New Game");
  text.push_back("Continue Game");
  text.push_back("Define Keys");
  text.push_back("High Scores");
  text.push_back("Fullscreen: ");
  text.push_back("Music: ");
  text.push_back("Sound Effects: ");
  text.push_back("Publish Scores: ");
  text.push_back("Help");
  text.push_back("Quit");
  active.resize(NUM_ITEMS);
  for (int i = 0; i < NUM_ITEMS; i++) {
    active[i] = true;
  }
  active[CONTINUE_GAME] = false;
  // TODO: change it to true after implementing the ShowHelp() function.
  active[HELP] = false;
  selected_item = NEW_GAME;
  starting_level = 1;
}

int Menu::GetStartingLevel() const
{
  return starting_level;
}

void Menu::EnableContinue()
{
  active[CONTINUE_GAME] = true;
  selected_item = CONTINUE_GAME;
}

void Menu::DisableContinue()
{
  active[CONTINUE_GAME] = false;
  selected_item = NEW_GAME;
}

void Menu::PrintItem(int i, Color color, const char* more)
{
  string txt = text[i];
  if (more) {
    txt += more;
  } else if (i == FULLSCREEN) {
    txt += fullscreen_on ? "ON" : "OFF";
  } else if (i == MUSIC) {
    txt += music_on ? "ON" : "OFF";
  } else if (i == SOUND_EFFECTS) {
    txt += sound_effects_on ? "ON" : "OFF";
  } else if (i == PUBLISH_SCORES) {
    txt += publish_scores ? "YES" : "NO";
  } else if (i == NEW_GAME) {
    stringstream ss;
    ss << " (Level " << starting_level << ")";
    txt += ss.str();
  }
  int w, h;
  PrintAt(menu_screen, menu_font, pos[i].x, pos[i].y, txt, color, &w, &h);
  item_height[i] = h;
  item_width[i] = w;
}

void Menu::PrintItem(int i, bool selected, const char* more)
{
  PrintItem(i, selected ? SELECTED_MENU_ITEM : MENU_ITEM, more);
}

void Menu::EraseItem(int i)
{
  if (menu_background_image) {
    SDL_Rect src, dest;
    src.x = dest.x = pos[i].x;
    src.y = dest.y = pos[i].y;
    src.w = item_width[i];
    src.h = item_height[i];
    SDL_BlitSurface(menu_background_image, &src, menu_screen, &dest);
  } else {
    ClearRect(menu_screen, MENU_BACKGROUND, pos[i].x, pos[i].y,
              item_width[i], item_height[i]);
  }
}

void Menu::Draw()
{
  if (menu_background_image) {
    SDL_BlitSurface(menu_background_image, NULL, menu_screen, NULL);
  } else {
    ClearRect(menu_screen, MENU_BACKGROUND, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
  }
  PrintAt(menu_screen, menu_title_font, 170, 50, "KARTOFEL", MENU_TITLE);
  int pos_x = 380;
  int pos_y = 280;
  pos.resize(NUM_ITEMS);
  // item_width and item_height are set inside PrintItem().
  item_width.resize(NUM_ITEMS);
  item_height.resize(NUM_ITEMS);
  for (int i = 0; i < NUM_ITEMS; i++) {
    if (!active[i]) {
      pos[i] = Point();
      continue;
    }
    Point p = { pos_x, pos_y };
    pos[i] = p;
    PrintItem(i, selected_item == i);
    pos_y += 30;
  }
  PrintAt(
    menu_screen, info_font, 50, 720,
    "\xc2\xa9 Copyright 2008 Pawe\xc5\x82 Aleksander Fedory\xc5\x84ski.  "
    "I dedicate this game to my beloved girlfriend Karen.",
    COPYRIGHT);
}

Menu::Choice Menu::Run()
{
  Draw();
  for (;;) {
    Update();
    int key;
    for (;;) {
      key = WaitForKey();
      if (key == SDLK_ESCAPE) {
        cout << "Bye.\n";
        Exit();
      }
      if (key == SDLK_LEFT || key == SDLK_RIGHT ||
          key == SDLK_UP || key == SDLK_DOWN || key == SDLK_RETURN) {
        break;
      }
    }
    if (key == SDLK_LEFT || key == SDLK_RIGHT) {
      // Change the starting level if we're on "New Game".
      if (selected_item == NEW_GAME) {
        if (key == SDLK_LEFT && starting_level > 1) {
          starting_level--;
        } else if (key == SDLK_RIGHT && starting_level < max_starting_level) {
          starting_level++;
        } else {
          continue;
        }
        PlaySound(SOUND_MENU_NAV);
        EraseItem(selected_item);
        PrintItem(selected_item, true);
      }
      continue;
    }
    if (key == SDLK_UP  ||  key == SDLK_DOWN) {
      PlaySound(SOUND_MENU_NAV);
      EraseItem(selected_item);
      PrintItem(selected_item, false);
      for (;;) {
        selected_item += key == SDLK_DOWN ? 1 : -1;
        if (selected_item >= NUM_ITEMS) {
          selected_item = 0;
        } else if (selected_item < 0) {
          selected_item = NUM_ITEMS - 1;
        }
        if (active[selected_item]) {
          break;
        }
      }
      PrintItem(selected_item, true);
      continue;
    }
    assert(key == SDLK_RETURN);
    if (selected_item != SOUND_EFFECTS) {
      PlaySound(SOUND_MENU_CLICK);
    }
    switch (selected_item) {
    case NEW_GAME:
      return CHOICE_NEW_GAME;
    case CONTINUE_GAME:
      return CHOICE_CONTINUE;
    case DEFINE_KEYS:
      DefineKeys();
      Draw();
      break;
    case HIGH_SCORES:
      high_scores.Show(menu_screen);
      Update();
      WaitForKey();
      Draw();
      break;
    case FULLSCREEN:
      EraseItem(FULLSCREEN);
      PrintItem(FULLSCREEN, true, fullscreen_on ? "OFF" : "ON");
      ToggleFullscreen();
      break;
    case MUSIC:
      EraseItem(MUSIC);
      PrintItem(MUSIC, true, music_on ? "OFF" : "ON");
      ToggleMusic();
      break;
    case SOUND_EFFECTS:
      EraseItem(SOUND_EFFECTS);
      PrintItem(SOUND_EFFECTS, true, sound_effects_on ? "OFF" : "ON");
      ToggleSoundEffects();
      PlaySound(SOUND_MENU_CLICK);
      break;
    case PUBLISH_SCORES:
      EraseItem(PUBLISH_SCORES);
      PrintItem(PUBLISH_SCORES, true,
                publish_scores ? "NO" : "YES");
      TogglePublishScores();
      PlaySound(SOUND_MENU_CLICK);
      break;
    case HELP:
      ShowHelp();
      break;
    case QUIT:
      cout << "Bye.\n";
      Exit();
    default:
      assert(!"Impossible");
    }
  }
}

void LoadLevels()
{
  for (int i = 0; i < NUM_LEVELS; i++) {
    stringstream filename;
    filename << LEVELS_FILENAME_PREFIX << i + 1 << ".txt";
    LoadLevel(filename.str(), &levels[i]);
  }
}

string CodeToUTF8(int code)
{
  char res[5] = { 0 };
  if (code <= 0x7f) {
    res[0] = code;
  } else if (code <= 0x7ff) {
    res[0] = 0xc0 | (code >> 6);
    res[1] = 0x80 | (code & 0x3f);
  } else if (code <= 0xffff) {
    res[0] = 0xe0 | (code >> 12);
    res[1] = 0x80 | ((code >> 6) & 0x3f);
    res[2] = 0x80 | (code & 0x3f);
  } else {
    res[0] = 0xf0 | (code >> 18);
    res[1] = 0x80 | ((code >> 12) & 0x3f);
    res[2] = 0x80 | ((code >> 6) & 0x3f);
    res[3] = 0x80 | (code & 0x3f);
  }
  return res;
}

void DrawAskNameBG()
{
  ClearRect(*current_screen, MSG_BORDER,
            ENTER_NAME_BG_X, ENTER_NAME_BG_Y,
            ENTER_NAME_BG_WIDTH, ENTER_NAME_BG_HEIGHT);
  ClearRect(*current_screen, ENTER_NAME_BACKGROUND,
            ENTER_NAME_BG_X + MSG_BORDER_WIDTH,
            ENTER_NAME_BG_Y + MSG_BORDER_WIDTH,
            ENTER_NAME_BG_WIDTH - 2 * MSG_BORDER_WIDTH,
            ENTER_NAME_BG_HEIGHT - 2 * MSG_BORDER_WIDTH);

  PrintAt(*current_screen, info_font,
          ENTER_NAME_BG_X + ENTER_NAME_PROMPT_X,
          ENTER_NAME_BG_Y + ENTER_NAME_PROMPT_Y,
          "Please enter your name:", ENTER_NAME_PROMPT);
  PrintAt(
    *current_screen, info_font_small,
    ENTER_NAME_BG_X + ENTER_NAME_NOTE_X,
    ENTER_NAME_BG_Y + ENTER_NAME_NOTE_Y,
    "If the Publish Scores option is set, your name may"
    " be submitted to the Kartofel Hall of Fame page.",
    ENTER_NAME_NOTE);
}

void EraseName()
{
  DrawAskNameBG();
}

bool PrintName(const string& name)
{
  return PrintAt(*current_screen, info_font,
                 ENTER_NAME_BG_X + ENTER_NAME_X,
                 ENTER_NAME_BG_Y + ENTER_NAME_Y,
                 name, ENTER_NAME, NULL, NULL,
                 MAX_NAME_WIDTH);
}

string AssembleName(const vector<string>& letters)
{
  string s;
  for (int i = 0; i < letters.size(); i++) {
    s += letters[i];
  }
  return s;
}

bool AskName(string* name)
{
  // static, so that we start from the name last entered.
  static vector<string> name_letters;
  DrawAskNameBG();
  int key;
  int unicode;
  for (;;) {
    EraseName();
    bool ok = PrintName(AssembleName(name_letters) + "_");
    if (!ok) {
      ok = PrintName(AssembleName(name_letters));
      if (!ok) {
        name_letters.pop_back();
        ok = PrintName(AssembleName(name_letters));
        assert(ok);
      }
    }
    Update();
    key = WaitForKey(&unicode);
    if (key == SDLK_DELETE || key == SDLK_BACKSPACE || key == SDLK_LEFT) {
      if (!name_letters.empty()) {
        name_letters.pop_back();
      }
    } else if (key == SDLK_RETURN) {
      PlaySound(SOUND_ENTER_NAME_RETURN);
      *name = AssembleName(name_letters);
      return true;
    } else if (key == SDLK_ESCAPE) {
      return false;
    } else if (unicode != 0) {
      name_letters.push_back(CodeToUTF8(unicode));
    } else {
      continue;
    }
  }
}

void SaveGameToFile()
{
  char filename[1000];
  snprintf(filename, sizeof(filename),
           "games/%d", static_cast<int>(time(NULL)));
  FILE* f = fopen(filename, "wb");
  if (!f) {
    cerr << "fopen(): " << filename << ": " << strerror(errno) << endl;
    exit(1);
  }
  string encoded;
  saved_game.Encode(&encoded);
  if (encoded.size() < fwrite(encoded.data(), encoded.size(), 1, f)) {
    cerr << "fwrite(): " << filename << ": " << strerror(errno) << endl;
    exit(1);
  }
  if (0 != fclose(f)) {
    cerr << "fclose(): " << filename << ": " << strerror(errno) << endl;
  }
}

struct ScheduleGameData {
  string name;
  int score;
  string encoded_game;
};

int SubmitGame(void* dataptr)
{
  ScheduleGameData* data = reinterpret_cast<ScheduleGameData*>(dataptr);
  SubmitGame(data->name, data->score, VERSION, data->encoded_game);
  delete data;
  return 0;
}

void ScheduleGameSubmission(const string& name, int score)
{
  ScheduleGameData* data = new ScheduleGameData;
  data->name = name;
  data->score = score;
  saved_game.Encode(&(data->encoded_game));
  SDL_Thread* thread = SDL_CreateThread(SubmitGame, data);
  if (!thread) {
    cerr << "SDL_CreateThread() failed: " << SDL_GetError() << endl;
    delete data;
    return;
  }
}

void MaybeScheduleGameSubmission(const string& name, int score)
{
  //SaveGameToFile();
  if (libcurl_successfully_initialized && publish_scores) {
    ScheduleGameSubmission(name, score);
  }
}

void Run()
{
  Init();
  LoadLevels();

  if(Mix_PlayMusic(music, -1) == -1) {
    cerr << "Mix_PlayMusic() failed: " << Mix_GetError() << endl;
    exit(-1);
  }
  
  Game game;
  Menu menu;
  string name;
  for (;;) {
    PlayGameResult game_res;
    current_screen = &menu_screen;
    AutoRepeatOn();
    Menu::Choice choice;
    for (;;) {
      choice = menu.Run();
      if (choice == Menu::CHOICE_NEW_GAME) {
        if (!AskName(&name)) {
          continue;
        }
      }
      break;
    }
    current_screen = &game_screen;
    AutoRepeatOff();
    switch (choice) {
    case Menu::CHOICE_NEW_GAME:
      game = Game();
      saved_game.Reset(menu.GetStartingLevel());
      game_res = PlayGame(&game, menu.GetStartingLevel());
      break;
    case Menu::CHOICE_CONTINUE:
      assert(game.Initialized());
      game_res = PlayGame(&game);
      break;
    default:
      assert(!"Impossible.");
    }
    if (game_res == GAME_PAUSE) {
      menu.EnableContinue();
    } else if (game_res == GAME_DEATH) {
      if (game.GetTimeLeft() <= 0) {
        WaitForKey();
      }
      MsgGameOver();
      Update();
      MaybeScheduleGameSubmission(name, game.GetScore());
      WaitForKey();
      menu.DisableContinue();
    } else if (game_res == GAME_SUCCESS) {
      MsgGameSuccess();
      Update();
      MaybeScheduleGameSubmission(name, game.GetScore());
      WaitForEnterOrEsc();
      menu.DisableContinue();
    }
    if ((game_res == GAME_DEATH || game_res == GAME_SUCCESS) &&
        high_scores.IsHighEnough(game.GetScore())) {
      high_scores.Add(name, game.GetScore());
      high_scores.Show(game_screen);
      Update();
      WaitForKey();
    }
  }
}

}  // namespace kartofel

int main(int argc, char** argv)
{
  kartofel::Run();
  return 0;
}
