//  This file is part of the computer game Kartofel.
//  Copyright (C) 2008 Pawel Aleksander Fedorynski <pfedor@fuw.edu.pl>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <iomanip>
#include <iostream>
#include <cassert>
#include <cmath>

#include "util.h"

namespace kartofel {

using namespace std;


const double PI = 3.14159265358979323846;

int DoubleToInt(double x)
{
  x += 0.5;
  return static_cast<int>(x > 0 ? x : x - 1);
}

void FindIntersections(int x0, int y0, int x1, int y1,
                       int xc, int yc, int r_int,
                       vector<Point>* res)
{
  // We are going to raise these to the power of 4 so long long int is needed.
  // Even with that, the parameters must be no greater than 10,000.
  long long int r = r_int;
  long long int x = x0 - xc;
  long long int y = y0 - yc;
  long long int u = x1 - x0;
  long long int w = y1 - y0;
  //cout << "x: " << x << ", y: " << y << ", u: " << u << ", w: " << w << endl;
  assert(u != 0 || w != 0);
  // Check whether there is an intersection at all.
  long long int n = x*u + y*w;
  long long int d = u*u + w*w;
  //cout << "n: " << n << ", d: " << d << endl;
  if (n <= 0  &&  d + n >= 0) {
/*
    double a = -1.0 * (x*u + y*w) / (u*u + w*w);  // -n/d
    if ((u*u + w*w) * a*a  +  2 * (x*u + y*w) * a  +  x*x + y*y  >  r*r) {
      return;
    }
*/
    if (d * (x*x + y*y) - n*n  >  d * r*r) {
      return;
    }
  } else {
    if (x*x + y*y > r*r  &&
        d + 2 * n + x*x + y*y > r*r) {
      return;
    }
  }
  // There is at least one intersection, solve the quadratic equation to
  // find it.
  long long int D = 4 * ((x*u + y*w) * (x*u + y*w) -
                         (u*u + w*w) * (x*x + y*y - r*r));
  //cout << "D: " << D << endl;
  assert(D >= 0);
  if (D == 0) {
    // There is exactly one solution.
    double a = -1.0 * (x*u + y*w) / (u*u + w*w);
    Point p;
    p.x = static_cast<int>(x0 + a * u);
    p.y = static_cast<int>(y0 + a * w);
    res->push_back(p);
  } else {
    // There are two solutions.
    double sqrtD = sqrt(D);
    double a1 = (-2 * (x*u + y*w) + sqrtD) / (2 * (u*u + w*w));
    if (a1 >= 0 && a1 <= 1) {
      Point p1;
      //cout << setprecision(20) << "p1: (" << x0 + a1 * u
      //     << ", " << y0 + a1 * w << ")\n";
      p1.x = DoubleToInt(x0 + a1 * u);
      p1.y = DoubleToInt(y0 + a1 * w);
      res->push_back(p1);
    }
    double a2 = (-2 * (x*u + y*w) - sqrtD) / (2 * (u*u + w*w));
    if (a2 >= 0 && a2 <= 1) {
      Point p2;
      //cout << setprecision(20) << "p2: (" << x0 + a2 * u
      //     << ", " << y0 + a2 * w << ")\n";
      p2.x = DoubleToInt(x0 + a2 * u);
      p2.y = DoubleToInt(y0 + a2 * w);
      // The rounding errors could have conflated the solutions.
      if (res->empty() || p2.x != ((*res)[0]).x  ||  p2.y != ((*res)[0]).y) {
        res->push_back(p2);
      }
    }
  }
}

inline long long int Det(long long int a11,
                         long long int a12,
                         long long int a21,
                         long long int a22)
{
  //cout << "Det(" << a11 << ", " << a12 << ", "
  //     << a21 << ", " << a22 << ")\n";
  return a11 * a22 - a12 * a21;
}

bool InBoundingBox(long long int x, long long int y, const LineSegment& ls)
{
  long long int x1 = ls.x1;
  long long int y1 = ls.y1;
  long long int x2 = ls.x2;
  long long int y2 = ls.y2;
  //cout << "InBoundingBox(" << x << ", " << y << ", [(" << x1 << ", "
  //     << y1 << "), (" << x2 << ", " << y2 << ")])\n";
  if (x1 > x2) {
    swap(x1, x2);
  }
  if (y1 > y2) {
    swap(y1, y2);
  }
  return x1 <= x && x <= x2 && y1 <= y && y <= y2;
}

bool LineSegmentsIntersect(const LineSegment& ls1,
                           const LineSegment& ls2,
                           int* intersection_x,
                           int* intersection_y)
{
  long long int x1 = ls1.x1;
  long long int y1 = ls1.y1;
  long long int x2 = ls1.x2;
  long long int y2 = ls1.y2;
  long long int x3 = ls2.x1;
  long long int y3 = ls2.y1;
  long long int x4 = ls2.x2;
  long long int y4 = ls2.y2;

  //cout << "x1: " << x1 << ", y1: " << y1 << ", x2: " << x2 << ", y2: " << y2
  //   << ", x3: " << x3 << ", y3: " << y3 << ", x4: " << x4 << ", y4: " << y4
  //   << endl;

  long long int d = Det(x1 - x2, y1 - y2, x3 - x4, y3 - y4);

  if (d == 0) {
    if (Det(x1 - x2, y1 - y2, x1 - x4, y1 - y4) != 0) {
      // The segments are parallel.
      return false;
    }
    // The segments are on the same line.
    if (InBoundingBox(x1, y1, ls2)) {
      // The starting point of the first segment is already in the second
      // segment.  Return it.
      *intersection_x = ls1.x1;
      *intersection_y = ls1.y1;
      return true;
    } else if (InBoundingBox(x2, y2, ls2)) {
      // The end point of the first segment is in the second segment.
      // Determine which end of the second segment is in the first segment
      // and return it.
      if (InBoundingBox(x3, y3, ls1)) {
        *intersection_x = ls2.x1;
        *intersection_y = ls2.y1;
        return true;
      } else {
        assert(InBoundingBox(x4, y4, ls1));
        *intersection_x = ls2.x2;
        *intersection_y = ls2.y2;
        return true;
      }
    } else {
      return false;
    }
  } else {
    //cout << "Not parallel.\n";
    long long int det1 = Det(x1, y1, x2, y2);
    long long int det2 = Det(x3, y3, x4, y4);
    long long int nx = Det(det1, x1 - x2, det2, x3 - x4);
    long long int ny = Det(det1, y1 - y2, det2, y3 - y4);
    if (x1 > x2) {
      swap(x1, x2);
    }
    if (y1 > y2) {
      swap(y1, y2);
    }
    if (x3 > x4) {
      swap(x3, x4);
    }
    if (y3 > y4) {
      swap(y3, y4);
    }
    //cout << "d: " << d << ", nx: " << nx << ", ny: " << ny << endl;
    // The condition below is equivalent to
    // InBoundingBox(nx / d, ny / d, [ (x1, y1), (x2, y2) ]) &&
    // InBoundingBox(nx / d, ny / d, [ (x3, y3), (x4, y4) ]), but doesn't
    // require going through floating point division, so we can be sure
    // that the answer to whether the line segments intersect is always
    // accurate.
    if (((d > 0  &&
          (d*x1 <= nx  &&  nx <= d*x2  &&  d*y1 <= ny  &&  ny <= d*y2 &&
           d*x3 <= nx  &&  nx <= d*x4  &&  d*y3 <= ny  &&  ny <= d*y4))  ||
         (d < 0  &&
          (d*x1 >= nx  &&  nx >= d*x2  &&  d*y1 >= ny  &&  ny >= d*y2 &&
           d*x3 >= nx  &&  nx >= d*x4  &&  d*y3 >= ny  &&  ny >= d*y4)))) {
      *intersection_x = DoubleToInt(1.0 * nx / d);
      *intersection_y = DoubleToInt(1.0 * ny / d);
      return true;
    } else {
      return false;
    }
  }
}

}  // namespace kartofel
